#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <time.h>
#include <unistd.h>

void client(char *path) {
  printf("client: running\n");
  int clientfd;
  int len = 0;
  struct sockaddr_un un;
  //  struct epoll_event *events = nullptr;
  //  event_t ev;
  int epollfd;

  clientfd = socket(AF_UNIX, SOCK_STREAM | SOCK_NONBLOCK | SOCK_CLOEXEC, 0);

  if (clientfd < 0) {
    printf("client: failed to create socket.%s\n", strerror(errno));
    exit(1);
  }
  printf("client: fd is: %d\n", clientfd);
  bzero(&un, sizeof(un));
  un.sun_family = AF_UNIX;
  strcpy(un.sun_path, path);
  len = strlen(un.sun_path) + sizeof(un.sun_family);

  /*
   * Create epoll context.
   */
  epollfd = epoll_create(64);

  if (epollfd < 0) {
    perror("client: epoll_create");
    exit(1);
  }

  struct epoll_event event;
  event.events = EPOLLIN | EPOLLOUT | EPOLLRDHUP | EPOLLET;
  event.data.ptr = nullptr;
  int status = epoll_ctl(epollfd, EPOLL_CTL_ADD, clientfd, &event);
  if (status == -1) {
    perror("client: epoll_ctl");
    exit(1);
  }

  if (connect(clientfd, (struct sockaddr *)&un, len) == -1) {
    perror("client: connect");
    exit(1);
  }
  printf("client: connected.\n");

  struct epoll_event events[16];
  for (;;) {
    int num = epoll_wait(epollfd, events, 16, 10);
    if (num == -1) {
      perror("client: epoll_wait");
      exit(1);
    }
    for (int i = 0; i < num; ++i) {
      if (events[i].events == 0) {
        continue;
      }
      printf("Events received: %d\n", events[i].events);
      if (events[i].events & EPOLLHUP) {
        printf("Got close!\n");
        close(clientfd);
        return;
      }
    }
  }
}

void server(char *path) {
  printf("server: running\n");
  int s, len;
  struct sockaddr_un local;  // remote;
                             //  char str[100];
                             //  int stat = 0;
                             //  int last_stat = 0;

  if ((s = socket(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0)) == -1) {
    perror("server: socket");
    exit(1);
  }

  local.sun_family = AF_UNIX;
  strcpy(local.sun_path, path);
  unlink(local.sun_path);
  len = strlen(local.sun_path) + sizeof(local.sun_family);

  if (bind(s, (struct sockaddr *)&local, len) == -1) {
    perror("server: bind");
    close(s);
    exit(1);
  }

  if (listen(s, 5) == -1) {
    perror("server: listen");
    close(s);
    exit(1);
  }

  for (;;) {
    printf("server: waiting for a connection...\n");
    struct sockaddr clientaddr;
    socklen_t addrlen = sizeof(clientaddr);
    int cs = accept(s, &clientaddr, &addrlen);
    if (cs == -1) {
      perror("server: accept");
      close(s);
      exit(1);
    }

    printf("server: connected with %d.\n", cs);
    close(cs);
    printf("server: closed fd %d\n", cs);
    break;
  }
}

int main(int argc, char *argv[]) {
  printf("\nmain: running\n");
  pthread_t client_thread, server_thread;

  int r = pthread_create(&server_thread, nullptr, (void *(*)(void *)) & server,
                         (void *)"socket1");
  if (r != 0) {
    perror("server create");
    exit(1);
  }

  sleep(2);
  r = pthread_create(&client_thread, nullptr, (void *(*)(void *)) & client,
                     (void *)"socket1");
  if (r != 0) {
    perror("client create");
    exit(1);
  }

  for (;;) {
    fflush(stdout);
  }

  return 0;
}
